# coding: utf-8

"""
Root Signals API

Root Signals JSON API provides a way to access Root Signals using provisioned API token

The version of the OpenAPI document: 1.0.0 (latest)
Generated by OpenAPI Generator (https://openapi-generator.tech)

Do not edit the class manually.
"""  # noqa: E501

from __future__ import annotations

import json
from enum import Enum

from typing_extensions import Self


class DocumentTypeEnum(str, Enum):
    """
    * `csv` - CSV * `pdf` - PDF * `html` - HTML * `docx` - DOCX * `doc` - DOC * `txt` - TXT * `md` - MD * `pptx` - PPTX * `ppt` - PPT * `xlsx` - XLSX * `xls` - XLS * `odt` - ODT * `rtf` - RTF * `xml` - XML * `eml` - EML * `epub` - EPUB
    """

    """
    allowed enum values
    """
    CSV = "csv"
    PDF = "pdf"
    HTML = "html"
    DOCX = "docx"
    DOC = "doc"
    TXT = "txt"
    MD = "md"
    PPTX = "pptx"
    PPT = "ppt"
    XLSX = "xlsx"
    XLS = "xls"
    ODT = "odt"
    RTF = "rtf"
    XML = "xml"
    EML = "eml"
    EPUB = "epub"

    @classmethod
    def from_json(cls, json_str: str) -> Self:
        """Create an instance of DocumentTypeEnum from a JSON string"""
        return cls(json.loads(json_str))
